//
//  TSCommonTool.swift
//  TSLiveWallpaper
//
//  Created by 100Years on 2024/12/20.
//

import Kingfisher
import Foundation
import AVFoundation
open class TSCommonTool {
    /// 下载图片，优先从缓存中取，支持设置缓存策略
    /// - Parameters:
    ///   - url: 图片的 URL 地址
    ///   - completion: 完成回调，返回图片或错误
    public static func fetchImageWithCache(
        from urlString: String,
        completion: @escaping (UIImage?, Error?) -> Void
    ) {
        if urlString.contains("http") == false,
           let image = UIImage(named: urlString){
            completion(image, nil)
            return
        }
        guard let url = URL(string: urlString) else{
            completion(nil, NSError(domain: "url null", code: 0))
            return
        }

        // 配置 Kingfisher 的缓存策略
        let options: KingfisherOptionsInfo = [
            .cacheOriginalImage,
            .memoryCacheExpiration(.days(7)),  // 内存缓存时间
            .diskCacheExpiration(.days(30)),  // 磁盘缓存时间
            .cacheSerializer(DefaultCacheSerializer.default)
        ]
        
        // 使用 Kingfisher 检查缓存是否存在
        let cache = ImageCache.default
        let cacheKey = url.absoluteString
        
        cache.retrieveImage(forKey: cacheKey, options: nil) { result in
            switch result {
            case .success(let value):
                if let image = value.image {
                    // 从缓存中取到图片
                    completion(image, nil)
                } else {
                    // 缓存中不存在图片，进行下载
                    downloadImage(from: url, options: options, completion: completion)
                }
            case .failure(let error):
                // 缓存检查失败，直接下载
                print("缓存检查失败: \(error.localizedDescription)")
                downloadImage(from: url, options: options, completion: completion)
            }
        }
    }
    
    
    /// 获取已缓存的磁盘图片文件路径
    /// - Parameters:
    ///   - urlString: 图片的 URL 地址
    ///   - completion: 完成回调，返回图片的磁盘路径或错误
    public static func fetchImagePathWithCache(
        from urlString: String,
        completion: @escaping (String?, Error?) -> Void
    ) {
        guard let url = URL(string: urlString) else {
            completion(nil, NSError(domain: "InvalidURL", code: 0, userInfo: [NSLocalizedDescriptionKey: "无效的 URL"]))
            return
        }
        
        let cache = ImageCache.default
        let cacheKey = url.absoluteString
        if cache.isCached(forKey: cacheKey) {
            completion(cache.cachePath(forKey: cacheKey), nil)
        }else {
          completion(nil, NSError(domain: "CachePathError", code: 1, userInfo: [NSLocalizedDescriptionKey: "无法获取缓存路径"]))
        }

    }
    

    /// 下载图片并缓存
    /// - Parameters:
    ///   - url: 图片的 URL 地址
    ///   - options: Kingfisher 下载和缓存的选项
    ///   - completion: 完成回调，返回图片或错误
    public static func downloadImage(
        from url: URL,
        options: KingfisherOptionsInfo,
        completion: @escaping (UIImage?, Error?) -> Void
    ) {
        KingfisherManager.shared.retrieveImage(with: url, options: options,progressBlock: { receivedSize, totalSize in
            debugPrint(receivedSize)
        }) { result in
            switch result {
            case .success(let value):
                completion(value.image, nil)
            case .failure(let error):
                completion(nil, error)
            }
        }
    }

}






public let kMainQueue = DispatchQueue.main
/// 主线程延迟执行回调
/// - Parameters:
///   - delay: 延迟时间（秒）
///   - completion: 延迟后的回调
public func kDelayOnMainThread(_ delay: TimeInterval, completion: @escaping () -> Void) {
    DispatchQueue.main.asyncAfter(deadline: .now() + delay) {
        completion()
    }
}

public func kDelayMainShort(completion: @escaping () -> Void) {
    DispatchQueue.main.asyncAfter(deadline: .now() + 0.1) {
        completion()
    }
}

/// 在主线程上执行代码
/// - Parameter block: 要执行的代码块
public func kExecuteOnMainThread(_ block: @escaping () -> Void) {
    if Thread.isMainThread {
        block()
    } else {
        DispatchQueue.main.async {
            block()
        }
    }
}


/// 子 size 根据容器 size,返回自己的最大 size
/// - Parameters:
///   - containerSize: 容器 Size
///   - subviewSize: 子 Size
///   - contentInset: contentInset
///   - contentMode: contentMode
/// - Returns: 最大 size
public func kCalculateScaledSize(containerSize: CGSize, subviewSize: CGSize, contentInset: UIEdgeInsets = .zero ,contentMode:UIView.ContentMode) -> CGSize {
    // 计算去除边距后的容器可用大小
    let availableWidth = containerSize.width - contentInset.left - contentInset.right
    let availableHeight = containerSize.height - contentInset.top - contentInset.bottom
    
    switch contentMode {
    case .scaleToFill:
        // 直接填充容器，忽略子视图比例
        return CGSize(width: availableWidth, height: availableHeight)
    case .scaleAspectFit:
        // 按比例适配容器，确保子视图完全显示在容器内
        let widthRatio = availableWidth / subviewSize.width
        let heightRatio = availableHeight / subviewSize.height
        let scale = min(widthRatio, heightRatio)
        return CGSize(width: subviewSize.width * scale, height: subviewSize.height * scale)
    case .scaleAspectFill:
        // 按比例填充容器，可能会裁剪子视图
        let widthRatio = availableWidth / subviewSize.width
        let heightRatio = availableHeight / subviewSize.height
        let scale = max(widthRatio, heightRatio)
        return CGSize(width: subviewSize.width * scale, height: subviewSize.height * scale)
    default:
        // 默认返回子视图原始大小
        return subviewSize
    }
}
